----------------------------------------------------------------------------------------------------
-- ExtendedAILights
----------------------------------------------------------------------------------------------------
-- Purpose: Specialization for extended AI lights to visualize vairous AI-States
--
-- Copyright (c) Creative-Mesh, 2021
----------------------------------------------------------------------------------------------------

ExtendedAILights = {}
ExtendedAILights.MOD_NAME = g_currentModName

function ExtendedAILights.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Lights, specializations) and 
    SpecializationUtil.hasSpecialization(AIVehicle, specializations) and 
    SpecializationUtil.hasSpecialization(Motorized, specializations)
end

function ExtendedAILights.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "setAILightState", ExtendedAILights.setAILightState)
end

function ExtendedAILights.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "loadSharedLight", ExtendedAILights.loadSharedLight)
end

function ExtendedAILights.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", ExtendedAILights)

    SpecializationUtil.registerEventListener(vehicleType, "onAIStart", ExtendedAILights)
    SpecializationUtil.registerEventListener(vehicleType, "onAIEnd", ExtendedAILights)

    SpecializationUtil.registerEventListener(vehicleType, "onStartMotor", ExtendedAILights)
    SpecializationUtil.registerEventListener(vehicleType, "onStopMotor", ExtendedAILights)
end

function ExtendedAILights:onLoad(savegame)
    if self.spec_extendedAILights == nil then
        self.spec_extendedAILights = self[string.format("spec_%s.extendedAILights", ExtendedAILights.MOD_NAME)]

        self.spec_extendedAILights.aiActiveLights = {}
        self.spec_extendedAILights.shaderAiActiveLights = {}
        self.spec_extendedAILights.aiInactiveLights = {}
        self.spec_extendedAILights.shaderAiInactiveLights = {}
    end
    local spec = self.spec_extendedAILights

    -- Load all AI active lights
    self:loadVisualLights(self.xmlFile, "vehicle.lights.aiLights.aiActiveLight", false, spec.aiActiveLights, spec.shaderAiActiveLights)

    -- Load all AI inactive lights
    self:loadVisualLights(self.xmlFile, "vehicle.lights.aiLights.aiInactiveLight", false, spec.aiInactiveLights, spec.shaderAiInactiveLights)
end

--Event Listener to: onStartMotor
--Changes AI Lights to AIState
function ExtendedAILights:onStartMotor()
    local spec = self.spec_extendedAILights

    local isActive = self:getIsAIActive()
    self:setAILightState(spec.aiInactiveLights, spec.shaderAiInactiveLights, not isActive)
    self:setAILightState(spec.aiActiveLights, spec.shaderAiActiveLights, isActive)
end

--Event Listener to: onStopMotor
--Changes AI Lights to off
function ExtendedAILights:onStopMotor()
    local spec = self.spec_extendedAILights

    self:setAILightState(spec.aiInactiveLights, spec.shaderAiInactiveLights, false)
    self:setAILightState(spec.aiActiveLights, spec.shaderAiActiveLights, false)
end

--Event Listener to: onAIStart
--Changes AI Lights to active
function ExtendedAILights:onAIStart()
    local spec = self.spec_extendedAILights

    self:setAILightState(spec.aiInactiveLights, spec.shaderAiInactiveLights, false)
    self:setAILightState(spec.aiActiveLights, spec.shaderAiActiveLights, true)
end

--Event Listener to: onAIEnd
--Changes AI Lights to inactive
function ExtendedAILights:onAIEnd()
    local spec = self.spec_extendedAILights

    self:setAILightState(spec.aiActiveLights, spec.shaderAiActiveLights, false)
    self:setAILightState(spec.aiInactiveLights, spec.shaderAiInactiveLights, true)
end

--Sets visibility of given light
--@param aiLights table
--@param isActive boolean
function ExtendedAILights:setAILightState(aiLights, shaderAiLights, isActive)
    if aiLights ~= nil and shaderAiLights ~= nil then
        -- Change state of visualLights
        for _, light in pairs(aiLights) do
            setVisibility(light.node, isActive)
        end

        -- Change state of shaderLight
        local dir = isActive and 1 or -1
        for _, shaderLight in ipairs(shaderAiLights) do
            local x,y,z,w = getShaderParameter(shaderLight.node, "lightControl")
            setShaderParameter(shaderLight.node, "lightControl", math.max(x+dir*shaderLight.intensity, 0), y, z, w, false)
            if shaderLight.toggleVisibility then
                setVisibility(shaderLight.node, isActive)
            end
        end
    end
end

--Overwritten function to load shared lights with ai lights
--@param superFunc function
--@param xmlFile integer
--@param key string
--@param sharedLight table
function ExtendedAILights:loadSharedLight(superFunc, xmlFile, key, sharedLight)
    local success = superFunc(self, xmlFile, key, sharedLight)

    if self.spec_extendedAILights == nil then
        self.spec_extendedAILights = self[string.format("spec_%s.extendedAILights", ExtendedAILights.MOD_NAME)]

        self.spec_extendedAILights.aiActiveLights = {}
        self.spec_extendedAILights.shaderAiActiveLights = {}
        self.spec_extendedAILights.aiInactiveLights = {}
        self.spec_extendedAILights.shaderAiInactiveLights = {}
    end

    if success and sharedLight.node ~= nil then
        local xmlFilename = getXMLString(xmlFile, key .. "#filename")
        if xmlFilename ~= nil then
            xmlFilename = Utils.getFilename(xmlFilename, self.baseDirectory)
            local lightXmlFile = loadXMLFile("sharedLight", xmlFilename)
            if lightXmlFile ~= nil and lightXmlFile ~= 0 then

                local function addLights(xml, key, targetTable)
                    local i = 0
                    while true do
                        local lightKey = string.format("light.%s(%d)", key, i)
                        if not hasXMLProperty(xml, lightKey) then
                            break
                        end
                        local node = I3DUtil.indexToObject(getParent(sharedLight.node), getXMLString(xml, lightKey.."#node"))
                        if node ~= nil then
                            if getHasShaderParameter(node, "lightControl") then
                                local intensity = getXMLFloat(xml, lightKey.."#intensity") or 25
                                local toggleVisibility = Utils.getNoNil(getXMLBool(xml, lightKey.."#toggleVisibility"), false)
                                if toggleVisibility then
                                    setVisibility(node, false)
                                end
                                table.insert(targetTable, {node=node, intensity=intensity, toggleVisibility=toggleVisibility})
                            else
                                print("Warning: Node '"..getName(node).."' has no shaderparameter 'lightControl'. Ignoring node!")
                            end
                        else
                            print("Warning: Could not find node for '"..key.."' in '"..xmlFilename.."'!")
                        end
                        i = i + 1
                    end
                end

                addLights(lightXmlFile, "aiActiveLight", self.spec_extendedAILights.shaderAiActiveLights)
                addLights(lightXmlFile, "aiInactiveLight", self.spec_extendedAILights.shaderAiInactiveLights)

                delete(lightXmlFile)
            end
        end
    end

    return success
end